;;;; -*- Mode: Lisp; indent-tabs-mode: nil -*-
;;;
;;; --- Wrapper over lfp_spawn(3)
;;;

#+xcvb (module (:depends-on ("ffi-types-unix")))

(in-package :iolib.os)

;;; Child process reaping protocol

;; Question: how do we survive locks across process dump & restore?
(defvar *child-process-lock*
  (bt:make-lock "lock for child process reaper"))
(defvar *waitpid-lock* ;; only one process should be using waitpid at a time.
  (bt:make-lock "lock for waitpid"))

(defmacro with-child-process-lock (() &body body)
  `(bt:with-lock-held (*child-process-lock*)
     ,@body))
(defmacro with-waitpid-lock (() &body body)
  `(bt:with-lock-held (*waitpid-lock*)
     ,@body))


(defvar *child-process-reaper-hook*
  ())

(defun register-child-reaper (reaper)
  "Register REAPER as a function to handle dead children processes.
REAPER must be a designator for a function taking two arguments and
returning a generalized boolean. The arguments are a PID and a STATUS
to be decoded with such functions as ISYS:WIFEXITED and ISYS:WEXITSTATUS.
The return value is true iff the reaper could indeed handle specified process."
  (with-child-process-lock ()
    (pushnew reaper *child-process-reaper-hook*)))

(defun unregister-child-reaper (reaper)
  (with-child-process-lock ()
    (setf *child-process-reaper-hook*
          (delete reaper *child-process-reaper-hook*))))

(defun reap-child (pid status)
  "Call each registered reaper until one is found that matches the child process."
  (with-child-process-lock ()
    (let ((hook *child-process-reaper-hook*))
      (if (null hook)
          (format *error-output*
                  "No registered reaper to take care of dead child, pid: ~D status: #x~X"
                  pid status)
          (let ((handled
                 (loop :for reaper :in hook :thereis (funcall reaper pid status))))
            (unless (or handled (not (and (zerop (isys:wifstopped status))
                                          (zerop (isys:wifcontinued status)))))
              (format *error-output*
                      "None of the registered reapers could handle dead child, pid: ~D status: #x~X"
                      pid status))
            handled)))))

(defun reap-children (&optional (options 0))
  (with-waitpid-lock ()
    (%reap-children options)))

(defun %reap-children (&optional (options 0))
  "For each child the status of which is available through waitpid
   (with given waitpid OPTIONS), invoke HANDLER with pid and status
   (to be decoded with isys::wifexited and friends).
   Return two values, a boolean and an integer.
   The first value is NIL if there are still active children, T if there are no more.
   The second value is a count of how many children were processed.
   NOTE1: this function must be run inside a WITH-WAITPID-LOCK.
   NOTE2: if this has any chance of running in a signal handler,
   then both WITH-WAITPID-LOCK and WITH-CHILD-PROCESS-LOCK need to
   be protected by some signal-blocking unwind-protect."
  (loop :for count :from 0 :do
    (multiple-value-bind (pid status)
        (handler-case
            (isys:waitpid -1 options)
          (isys:echild () ;; No more children
            (return (values t count))))
      (when (zerop pid) ;; More active children
        (return (values nil count)))
      (reap-child pid status)
      (setf options (logior options isys:wnohang)))))

;;; Process

(defclass process ()
  ((pid    :initarg :pid :reader process-pid)
   (status :initform nil :reader process-status) ; exit status. nil if not reaped yet.
   (stdin  :initform nil :reader process-stdin)
   (stdout :initform nil :reader process-stdout)
   (stderr :initform nil :reader process-stderr)
   (fdmap  :initarg :fdmap :reader process-fdmap)))

(defun process-stdin-fd (process)
  (values (gethash +stdin+ (process-fdmap process))))
(defun process-stdout-fd (process)
  (values (gethash +stdout+ (process-fdmap process))))
(defun process-stderr-fd (process)
  (values (gethash +stderr+ (process-fdmap process))))

(defmethod initialize-instance :after ((process process) &key stdin stdout stderr)
  (let ((stdin (or stdin (process-stdin-fd process)))
        (stdout (or stdout (process-stdout-fd process)))
        (stderr (or stderr (process-stderr-fd process))))
    (with-slots ((in stdin) (out stdout) (err stderr))
        process
      (setf in  (and stdin  (make-instance 'iolib.streams:dual-channel-gray-stream :fd stdin))
            out (and stdout (make-instance 'iolib.streams:dual-channel-gray-stream :fd stdout))
            err (and stderr (make-instance 'iolib.streams:dual-channel-gray-stream :fd stderr))))))

(defmethod close ((process process) &key abort)
  (with-slots (pid status stdin stdout stderr)
      process
    (when stdin  (close stdin  :abort abort))
    (when stdout (close stdout :abort abort))
    (when stderr (close stderr :abort abort))
    (unless status
      (process-wait process))
    (setf pid nil stdin nil stdout nil stderr nil)))

(defun tty-read-fn (fd buf nbytes)
  (handler-case
      (isys:read fd buf nbytes)
    (isys:eio () 0)))

(defun tty-write-fn (fd buf nbytes)
  (handler-case
      (isys:write fd buf nbytes)
    (isys:eio ()
      (error 'isys:epipe
             :handle fd
             :syscall "write"))))

(defclass tty-stream (iolib.streams:dual-channel-gray-stream)
  ()
  (:default-initargs :read-fn  #'tty-read-fn
                     :write-fn #'tty-write-fn))

(defclass process ()
  ((pid    :initarg :pid :reader process-pid)
   (status :initform :running)
   (closed :initform nil)
   (stdin  :reader process-stdin)
   (stdout :reader process-stdout)
   (stderr :reader process-stderr)))

(defmethod initialize-instance :after ((process process) &key
                                       stdin stdout stderr external-format)
  (with-slots ((in stdin) (out stdout) (err stderr))
      process
    (when stdin
      (setf in  (make-instance 'tty-stream :fd stdin
                               :external-format external-format)))
    (when stdout
      (setf out (make-instance 'tty-stream :fd stdout
                               :external-format external-format)))
    (when stderr
      (setf err (make-instance 'tty-stream :fd stderr
                               :external-format external-format)))))

(defmethod close ((process process) &key abort)
  (if (slot-value process 'closed)
      nil
      (macrolet ((close-process-stream (slot)
                   `(when (slot-boundp process ',slot)
                      (close (slot-value process ',slot) :abort abort)
                      (slot-makunbound process ',slot))))
        (close-process-stream stdin)
        (close-process-stream stdout)
        (close-process-stream stderr)
        (process-status process :wait (not abort))
        (setf (slot-value process 'closed) t)
        t)))
>>>>>>> 03fb1dcd7b50621cdbbd3f21ad1e82b090c3f69b

(defmethod print-object ((o process) s)
  (print-unreadable-object (o s :type t :identity t)
    (format s "~S ~S ~S ~S"
            :pid (process-pid o)
            :status (process-status o))))

(defun exit-status (status)
  (cond
    ((isys:wifexited status)
     (isys:wexitstatus status))
    ((isys:wifsignaled status)
     (values (isys:wtermsig* status)
             (isys:wcoredump status)))))

(defmethod process-status ((process process) &key wait)
  (if (integerp (slot-value process 'status))
      (exit-status (slot-value process 'status))
      (multiple-value-bind (pid status)
          (isys:waitpid (process-pid process)
                        (if wait 0 isys:wnohang))
        (cond
          ((zerop pid)
           :running)
          (t
           (setf (slot-value process 'status) status)
           (exit-status status))))))

(defmethod process-activep ((process process))
  (eql :running (process-status process)))

(defmethod process-kill ((process process) &optional (signum :sigterm))
  (isys:kill (process-pid process) signum)
  process)


(defmacro with-lfp-spawn-arguments ((attributes file-actions pid) &body body)
  (with-gensyms (spawnattr-initialized-p file-actions-initialized-p)
    `(with-foreign-objects ((,attributes 'lfp-spawnattr-t)
                            (,file-actions 'lfp-spawn-file-actions-t)
                            (,pid 'pid-t))
       (let ((,spawnattr-initialized-p nil)
             (,file-actions-initialized-p nil))
         (unwind-protect
              (progn
                (setf ,spawnattr-initialized-p
                      (lfp-spawnattr-init ,attributes))
                (setf ,file-actions-initialized-p
                      (lfp-spawn-file-actions-init ,file-actions))
                ,@body)
           (when ,spawnattr-initialized-p
             (lfp-spawnattr-destroy ,attributes))
           (when ,file-actions-initialized-p
             (lfp-spawn-file-actions-destroy ,file-actions)))))))

(defun allocate-argv (argv program arglist)
  ;; copy program name
  (setf (mem-aref argv :pointer 0)
        (foreign-string-alloc program))
  ;; copy program arguments
  (loop :for i :from 1
        :for arg :in arglist :do
        (setf (mem-aref argv :pointer i)
              (foreign-string-alloc arg))))

(defvar *default-shell* "/bin/sh")

(defun find-program (program)
  (cond
    ((eql :shell program)
     (or (nth-value 6 (isys:getpwuid (isys:getuid)))
         *default-shell*))
    (t
     (file-path-namestring program))))

(defmacro with-argv (((arg0 argv) program arguments) &body body)
  (with-gensyms (argc)
    `(let ((,program (find-program ,program))
           (,argc (+ 2 (length ,arguments))))
       (with-foreign-object (,argv :pointer ,argc)
         (isys:bzero ,argv (* ,argc (isys:sizeof :pointer)))
         (unwind-protect
              (progn
                (allocate-argv ,argv ,program ,arguments)
                (let ((,arg0 (mem-ref ,argv :pointer)))
                  ,@body))
           (delocate-null-ended-list ,argv))))))

(defun effect-redirections (file-actions spec &key (close-other-fds t))
  "SPEC is a list of (FD . REDIR) commands.
Returns a hash-table mapping subprocess FD to results."
  (let ((results (make-hash-table :test 'equal)) ; results
        (shuffle (make-hash-table :test 'equal)) ; mapping from child fd to parent fd
        (unshuffle (make-hash-table :test 'equal)) ; mapping from parent fd to child fd
        (parent-close (make-hash-table :test 'equal)) ; things to do on the parent afterwards
        (dups (make-hash-table :test 'equal)) ; duplicates after redirections
        (open-max (isys:open-max))) ; total number of fd's
    (loop :for (fd . redir) :in spec :do
      (check-type fd unsigned-byte)
      (assert (< fd open-max))
      (check-type redir (or symbol unsigned-byte cons))
      (when (integerp redir) (assert (< redir open-max)))
      (effect-redirection results shuffle unshuffle dups parent-close fd redir))
    ;; close all fd's that haven't been explicitly redirected
    ;; TODO: use introspection via /proc/self/fd to minimize number of closes?
    (loop :for fd :being :the :hash-keys :of dups :using (:hash-value dup) :do
      (assert (gethash dup shuffle)))
    (when close-other-fds
      (loop :for fd :below open-max
        :when (fd-open-p fd)
        :unless (nth-value 1 (gethash fd shuffle)) :do
        (setf (gethash fd shuffle) nil)))
    (loop
      :with child-fds = (loop :for fd :being :the :hash-keys :of shuffle :collect fd)
      :for fd :in child-fds
      :when (nth-value 1 (gethash fd shuffle)) :do
      (let ((basefd
             (loop
               ;; The linearity of redirections, enforced in effect-redirection,
               ;; ensures that any circularity will be an actual circle and goes
               ;; back to the starting point. No loops that start at one ancestor of it,
               ;; no multiple loops that share multiple nodes.
               :for prev = fd :then xfd
               :for xfd = (gethash prev unshuffle) :do
               (cond
                 ((null xfd)
                  (return prev))
                 ((= fd xfd)
                  (return nil))))))
        (cond
          ((equal fd (gethash fd shuffle))
           (remhash fd shuffle))
          (t
           (when (null basefd) ; break circularity by duplicating a fd
             (let ((tempfd (isys:dup fd))
                   (xfd (gethash fd unshuffle)))
               (setf (gethash xfd shuffle) tempfd)
               (setf (gethash tempfd unshuffle) xfd)
               (setf (gethash tempfd parent-close) nil)
               (setf basefd fd)))
           (loop :for targetfd = basefd :then source
             :for source = (prog1 (gethash targetfd shuffle) (remhash targetfd shuffle)) :do
             (etypecase source
               (unsigned-byte
                (unless (= source targetfd)
                  (posix-spawn-file-actions-adddup2 file-actions source targetfd)))
               (null
                (posix-spawn-file-actions-addclose file-actions targetfd)
                (return))
               (cons
                (destructuring-bind (path flags mode) source
                  (check-type path string)
                  (check-type flags integer)
                  (check-type mode integer)
                  (posix-spawn-file-actions-addopen file-actions targetfd path flags mode))
                (return))))))))
    (loop :for targetfd :being :the :hash-keys :of dups :using (:hash-value source) :do
      (posix-spawn-file-actions-adddup2 file-actions source targetfd))
    (values results parent-close)))

(defun normalize-open-flags (oflags &optional mode)
  (etypecase oflags
    (unsigned-byte
     oflags)
    ((member nil :input)
     isys:o-rdonly)
    ((eql :output)
     (logior isys:o-wronly (if mode isys:o-creat 0)))
    ((eql :append)
     (logior isys:o-wronly isys:o-append (if mode isys:o-creat 0)))
    ((eql :io)
     (logior isys:o-rdwr (if mode isys:o-creat 0)))))

(defun normalize-open-mode (mode)
  (etypecase mode
    (unsigned-byte
     mode)
    (null
     0)
    (t
     #o666)))

(defun ensure-hash (key table value)
  (assert (not (nth-value 1 (gethash key table))))
  (setf (gethash key table) value))

(defun effect-open-redirection (shuffle unshuffle parent-close
                                fd path flags mode)
  (declare (ignorable unshuffle parent-close))
  ;; #+trust-addopen (ensure-hash fd shuffle (list path flags mode))
  ;; #-trust-addopen
  (let ((redir (isys:open path flags mode)))
    (ensure-hash fd shuffle redir)
    (unless (= fd redir)
      (ensure-hash redir shuffle nil))
    (ensure-hash redir unshuffle fd)
    (ensure-hash redir parent-close nil)))

(defun effect-redirection (results shuffle unshuffle dups parent-close fd redir)
  (etypecase redir
    (unsigned-byte
     (ensure-hash fd shuffle redir)
     (ensure-hash redir unshuffle fd))
    (iolib.streams:dual-channel-fd-mixin
     (let ((rfd (iolib.streams:fd-of redir)))
       (ensure-hash fd shuffle rfd)
       (ensure-hash rfd unshuffle fd)))
    (null
     (ensure-hash fd shuffle nil))
    ((eql :stdout)
     (ensure-hash fd shuffle nil)
     (ensure-hash fd dups 1))
    ((eql :null)
     (effect-open-redirection
      shuffle unshuffle parent-close
      fd "/dev/null" isys:o-rdwr 0))
    ((eql :null-in)
     (effect-open-redirection
      shuffle unshuffle parent-close
      fd "/dev/null" isys:o-rdonly 0))
    ((eql :null-out)
     (effect-open-redirection
      shuffle unshuffle parent-close
      fd "/dev/null" isys:o-wronly 0))
    (cons ; path open-flag open-mode
     (destructuring-bind (path &optional oflags mode) redir
       (effect-open-redirection
        shuffle unshuffle parent-close
        fd (file-path-namestring path)
        (normalize-open-flags oflags mode)
        (normalize-open-mode mode))))
    ((eql t)
     (ensure-hash fd shuffle fd)
       (ensure-hash fd unshuffle fd))
      ((eql :pipe-in)
       (multiple-value-bind (read-fd write-fd) (isys:pipe)
         (ensure-hash fd shuffle read-fd)
         (ensure-hash read-fd unshuffle fd)
         (ensure-hash fd results write-fd)
         (ensure-hash read-fd parent-close nil)))
      ((eql :pipe-out)
       (multiple-value-bind (read-fd write-fd) (isys:pipe)
         (ensure-hash fd shuffle write-fd)
         (ensure-hash write-fd unshuffle fd)
         (ensure-hash fd results read-fd)
         (ensure-hash write-fd parent-close nil))))
  (values))

(defun effect-signal-attributes (attributes sigmask sigdefault)
  (when sigmask
    (with-foreign-object (%set 'isys:sigset-t)
      (isys:encode-sigset %set sigmask)
      (posix-spawnattr-getsigmask attributes %set)))
  (when sigdefault
    (with-foreign-object (%set 'isys:sigset-t)
      (isys:encode-sigset %set sigdefault)
      (posix-spawnattr-getsigdefault attributes %set)))
  (values))

(defun spawn-process (program arguments &key (search t) environment redirect
                      sigmask sigdefault)
  (with-posix-spawn-arguments (attributes file-actions pid)
    (with-argv ((arg0 argv) program arguments)
      (with-c-environment (env environment)
        (multiple-value-bind (results to-be-closed)
            (effect-redirections file-actions redirect)
          (effect-signal-attributes attributes sigmask sigdefault)
          (funcall (if search 'posix-spawnp 'posix-spawn)
                   pid arg0 file-actions attributes argv env)
          (loop :for fd :being :the :hash-keys :of to-be-closed :do (isys:close fd))
          (values (mem-ref pid 'pid-t) results))))))

(defvar *children-processes* (make-hash-table :test 'equal))

(defun create-process-child-reaper (pid status)
  (when (and (zerop (isys:wifstopped status)) (zerop (isys:wifcontinued status)))
    (let ((process (gethash pid *children-processes*)))
      (when process
        (remhash pid *children-processes*)
        (with-slots ((xpid pid) (xstatus status)) process
          (setf xpid nil)
          (setf xstatus status))
        t))))

(defun create-process (program &optional arguments
                       &key (search t) (environment t)
                       redirect (stdin t) (stdout t) (stderr t)
                       sigmask sigdefault)
  ;; program: :shell - the current user's shell
  ;;          file-path designator - a path
  ;; arguments: list of strings
  ;; search: boolean. whether or not to search PROGRAM in $PATH when PROGRAM only names a file,
  ;;                  i.e., it's a relative path whose directory is NIL
  ;; environment: t - inherit environment
  ;;              nil - NULL environment
  ;;              alist - the environment to use
  ;; redirect: an alist mapping fd number for the child to redirection
  ;;   each redirection may be
  ;;         file-descriptor designator(integer or stream) - file descriptor, redirect to it
  ;;         :pipe-in :pipe-out - create pipe, redirect the child descriptor to one end
  ;;                and wrap the other end into a stream which goes into PROCESS slot.
  ;;         :null :null-in :null-out - redirect to bit bucket, rdwr, rdonly, wronly.
  ;;         t - inherit
  ;;         nil - close
  ;; stdin, stdout, stderr: a redirection, with additional autodetection
  ;;         file-path designator, :null, :pipe - redirect with the standard direction
  ;;             (creating file if output direction and it doesn't exist)
  ;;         :stdout - for stderr, indicates it must copy stdout.
  (register-child-reaper 'create-process-child-reaper)
  (flet ((normalize-input-redirection (x)
           (etypecase x
             ((eql :pipe) :pipe-in)
             ((eql :null) :null-in)
             ((or cons unsigned-byte symbol) x)
             ((or string pathname) (list x :input))))
         (normalize-output-redirection (x)
           (etypecase x
             ((eql :pipe) :pipe-out)
             ((eql :null) :null-out)
             ((or cons unsigned-byte symbol) x)
             ((or string pathname) (list x :output t)))))
    (with-child-process-lock ()
      (multiple-value-bind (pid fdmap)
          (spawn-process
           program arguments
           :search search :environment environment
           :redirect (append
                      (when stdin (list (cons +stdin+ (normalize-input-redirection stdin))))
                      (when stdout (list (cons +stdout+ (normalize-output-redirection stdout))))
                      (when stderr (list (cons +stderr+ (normalize-output-redirection stderr))))
                      redirect)
           :sigmask sigmask :sigdefault sigdefault)
        (let ((process (make-instance 'process :pid pid :fdmap fdmap)))
          (setf (gethash pid *children-processes*) process)
          process)))))

#|
(defun redirect-one-stream (file-actions fd stream &optional flags (mode #o644) close-old-fd)
  (flet ((dup-from-path (path)
           (lfp-spawn-file-actions-addopen file-actions fd path flags mode))
         (dup-from-fd (oldfd)
           (lfp-spawn-file-actions-adddup2 file-actions oldfd fd)
           (when close-old-fd
             (lfp-spawn-file-actions-addclose file-actions oldfd))))
    (etypecase stream
      ((eql t) nil)
      ((or string file-path pathname)
       (dup-from-path (file-path-namestring stream)))
      ((eql :null)
       (dup-from-path "/dev/null"))
      (unsigned-byte
       (dup-from-fd stream))
      (iolib.streams:dual-channel-fd-mixin
       (dup-from-fd (iolib.streams:fd-of stream)))
      (null
       (lfp-spawn-file-actions-addclose file-actions fd)))))

(defun redirect-to-pipes (file-actions fd keep-write-fd)
  (multiple-value-bind (pipe-parent pipe-child)
      (isys:pipe)
    (when keep-write-fd (rotatef pipe-parent pipe-child))
    (lfp-spawn-file-actions-adddup2 file-actions pipe-child fd)
    (lfp-spawn-file-actions-addclose file-actions pipe-parent)
    (lfp-spawn-file-actions-addclose file-actions pipe-child)
    (values pipe-parent pipe-child)))

(defun setup-redirections (file-actions stdin stdout stderr ptmfd pts)
  (let (infd infd-child outfd outfd-child errfd errfd-child)
    ;; Standard input
    (case stdin
      (:pipe
       (setf (values infd infd-child)
             (redirect-to-pipes file-actions +stdin+ t)))
      (:pty
       (setf infd (isys:dup ptmfd))
       (redirect-one-stream file-actions +stdin+ pts isys:o-rdonly))
      (t (redirect-one-stream file-actions +stdin+ stdin isys:o-rdonly)))
    ;; Standard output
    (case stdout
      (:pipe
       (setf (values outfd outfd-child)
             (redirect-to-pipes file-actions +stdout+ nil)))
      (:pty
       (setf outfd (isys:dup ptmfd))
       (redirect-one-stream file-actions +stdout+ pts (logior isys:o-wronly
                                                              isys:o-creat)))
      (t (redirect-one-stream file-actions +stdout+ stdout (logior isys:o-wronly
                                                                   isys:o-creat))))
    ;; Standard error
    (case stderr
      (:pipe
       (setf (values errfd errfd-child)
             (redirect-to-pipes file-actions +stderr+ nil)))
      (:pty
       (setf errfd (isys:dup ptmfd))
       (redirect-one-stream file-actions +stderr+ pts (logior isys:o-wronly
                                                              isys:o-creat)))
      (t (redirect-one-stream file-actions +stderr+ stderr (logior isys:o-wronly
                                                                   isys:o-creat))))
    (values infd infd-child outfd outfd-child errfd errfd-child)))

(defun close-fds (&rest fds)
  (dolist (fd fds)
    (when fd (isys:close fd))))

(defun setup-slave-pty ()
  (let ((ptmfd (isys:openpt (logior isys:o-rdwr isys:o-noctty isys:o-cloexec))))
    (isys:grantpt ptmfd)
    (isys:unlockpt ptmfd)
    (values ptmfd (isys:ptsname ptmfd))))

(defmacro with-pty ((ptmfd pts) &body body)
  `(multiple-value-bind (,ptmfd ,pts)
       (setup-slave-pty)
     (unwind-protect
          (locally ,@body)
       (close-fds ,ptmfd))))

(defmacro with-redirections (((infd outfd errfd)
                              (file-actions stdin stdout stderr))
                             &body body)
  (with-gensyms (infd-child outfd-child errfd-child ptmfd pts)
    `(with-pty (,ptmfd ,pts)
       (multiple-value-bind (,infd ,infd-child ,outfd ,outfd-child ,errfd ,errfd-child)
           (setup-redirections ,file-actions ,stdin ,stdout ,stderr ,ptmfd ,pts)
         (unwind-protect-case ()
             (locally ,@body)
           (:always
            (close-fds ,infd-child ,outfd-child ,errfd-child))
           (:abort
            (close-fds ,infd ,outfd ,errfd)))))))

(defun process-other-spawn-args (attributes new-session current-directory
                                 uid gid resetids)
  (when new-session
    (lfp-spawnattr-setsid attributes))
  (when current-directory
    (lfp-spawnattr-setcwd attributes current-directory))
  (when uid
    (lfp-spawnattr-setuid attributes uid))
  (when gid
    (lfp-spawnattr-setgid attributes gid))
  (when resetids
    (lfp-spawnattr-setflags attributes lfp-spawn-resetids)))

;; program: :shell - the system shell
;;          file-path designator - a path
;; arguments: list
;; environment: t - inherit environment
;;              nil - NULL environment
;;              alist - the environment to use
;; stdin, stdout, stderr:
;;         file-path designator - open file, redirect to it
;;         :null - redirect to /dev/null - useful because /dev/null doesn't exist on Windows
;;         file-descriptor designator(integer or stream) - file descriptor, redirecto to it
;;         :pipe - create pipe, redirect the child descriptor to one end and wrap the other end
;;                 into a stream which goes into PROCESS slot
;;         t - inherit
;;         nil - close
;; new-session: boolean - create a new session using setsid()
;; current-directory: path - a directory to switch to before executing
;; uid: user id - unsigned-byte or string
;; gid: group id - unsigned-byte or string
;; resetids: boolean - reset effective UID and GID to saved IDs

(defun create-process (program-and-args &key (environment t)
                       (stdin :pipe) (stdout :pipe) (stderr :pipe)
                       new-session current-directory uid gid resetids
                       (external-format :utf-8))
  (flet ((new-ctty-p (stdin stdout stderr)
           (or (eql :pty stdin)
               (eql :pty stdout)
               (eql :pty stderr))))
    (destructuring-bind (program &rest arguments)
        (ensure-list program-and-args)
      (when (new-ctty-p stdin stdout stderr)
        (setf new-session t))
      (with-argv ((arg0 argv) program arguments)
        (with-c-environment (envp environment)
          (with-lfp-spawn-arguments (attributes file-actions pid)
            (with-redirections ((infd outfd errfd)
                                (file-actions stdin stdout stderr))
              (process-other-spawn-args attributes new-session current-directory
                                        uid gid resetids)
              (lfp-spawnp pid arg0 argv envp file-actions attributes)
              (make-instance 'process :pid (mem-ref pid 'pid-t)
                             :stdin infd :stdout outfd :stderr errfd
                             :external-format external-format))))))))
|#

(defun run-program (program-and-args &key (environment t) (stderr :pipe)
                    (external-format :utf-8))
  (flet ((slurp (stream)
           (with-output-to-string (s)
             (loop :for c := (read-char stream nil nil)
                   :while c :do (write-char c s)))))
    (let ((process (create-process program-and-args
                                   :environment environment
                                   :stdin nil
                                   :stdout :pipe
                                   :stderr stderr
                                   :external-format external-format)))
      (unwind-protect
           (values (process-status process :wait t)
                   (slurp (process-stdout process))
                   (if (eql :pipe stderr)
                       (slurp (process-stderr process))
                       (make-string 0)))
        (close process)))))
<<<<<<< HEAD

(defmethod process-wait ((process process))
  (with-waitpid-lock ()
    (loop :until (process-status process) :do
      (%reap-children)))
  (process-status process))

(defmethod process-kill ((process process) &optional (signum :sigterm))
  (isys:kill (process-pid process) signum))
=======
>>>>>>> 03fb1dcd7b50621cdbbd3f21ad1e82b090c3f69b
